MusicianOptions = LibStub("AceAddon-3.0"):NewAddon("MusicianOptions", "AceEvent-3.0")

local MusicianButtonGetMenu
local MusicianKeyboardNotekey

local MusicianLiveSetSustain
local MusicianLiveNoteOn
local MusicianLiveNoteOff
local MusicianLiveAllNotesOff

local isSustained = false
local useGW2 = false

-- Recreate some Musician.Live variables that are required
local STREAM_PADDING = 10 -- Number of seconds to wait before ending streaming without activity

local NOTEON = {}
NOTEON.HANDLE = 1
NOTEON.LAYER = 2
NOTEON.KEY = 3
NOTEON.INSTRUMENT = 4
NOTEON.IS_CHORD_NOTE = 5
NOTEON.SOURCE = 6

local notesOn = {}
local sustainedNotes = {}
local isPlayingLive = false
local playingLiveTimer

--- muteGameMusic
--
local function muteGameMusic()
	isPlayingLive = true
	Musician.Utils.MuteGameMusic()
	if playingLiveTimer then
		playingLiveTimer:Cancel()
	end
end

--- unmuteGameMusic
--
local function unmuteGameMusic()
	if playingLiveTimer then
		playingLiveTimer:Cancel()
	end
	playingLiveTimer = C_Timer.NewTimer(STREAM_PADDING, function()
		isPlayingLive = false
		Musician.Utils.MuteGameMusic()
		playingLiveTimer = nil
	end)
end


--- OnInitialize
function MusicianOptions:OnInitialize()
	-- Check for outdated Musician version
	if Musician.Utils.VersionCompare(MusicianOptions.MUSICIAN_MIN_VERSION, GetAddOnMetadata("Musician", "Version")) > 0 then
		C_Timer.After(10, function()
			Musician.Utils.Error(MusicianOptions.Msg.ERR_OUTDATED_MUSICIAN_VERSION)
		end)
		return
	end
	MusicianOptions.OnReady()
end

--- Second initialization phase
function MusicianOptions.OnReady()
	-- Is MusicianList already hooked? Hook to MusicianList instead
	if type(GetAddOnMetadata("MusicianList", "Version")) == "string" then
		-- Ensure we hook after MusicianList has hooked to MusicianButton
		C_Timer.After(1, function()
			MusicianButtonGetMenu = MusicianList.GetMenu
			MusicianButton.GetMenu = MusicianOptions.GetMenu
		end)
	else -- Musician List is not installed, it's okay to hook to base function
		MusicianButtonGetMenu = MusicianButton.GetMenu
		MusicianButton.GetMenu = MusicianOptions.GetMenu
	end

	MusicianKeyboardNotekey = MusicianKeyboard.NoteKey

	-- Hook all the live functions, because the notesOn and sustainedNotes variables are local.
	-- Create new local variables and put in our own functions instead.
	MusicianLiveSetSustain = Musician.Live.SetSustain
	MusicianLiveNoteOn = Musician.Live.NoteOn
	MusicianLiveNoteOff = Musician.Live.NoteOff
	MusicianLiveAllNotesOff = Musician.Live.AllNotesOff
end

-- Have to copy these functions due to new local variables.
-- Otherwise the change won't work.
function MusicianOptions.NoteKey(down, keyValue)
	local layer = 0
	local noteKey = 0
	Gw2Keys = {}
	Gw2Keys["1"] = "KeyQ"
	Gw2Keys["2"] = "KeyW"
	Gw2Keys["3"] = "KeyE"
	Gw2Keys["4"] = "KeyR"
	Gw2Keys["5"] = "KeyT"
	Gw2Keys["6"] = "KeyY"
	Gw2Keys["7"] = "KeyU"
	Gw2Keys["8"] = "KeyI"

	Gw2Keys["NUMPAD1"] = "KeyQ"
	Gw2Keys["NUMPAD2"] = "KeyW"
	Gw2Keys["NUMPAD3"] = "KeyE"
	Gw2Keys["NUMPAD4"] = "KeyR"
	Gw2Keys["NUMPAD5"] = "KeyT"
	Gw2Keys["NUMPAD6"] = "KeyY"
	Gw2Keys["NUMPAD7"] = "KeyU"
	Gw2Keys["NUMPAD8"] = "KeyI"

	local key = Gw2Keys[keyValue]
	local note = Musician.Keyboard.mapping[key]

	--Octave switching, just like in Gw2.
	if note == nil then
		if keyValue == "9" or keyValue == "NUMPADPLUS" or keyValue == "X" then 
			if down then
				Musician.Keyboard.ShiftKeys(layer, 14)
			end
			return true
		elseif keyValue == "0" or keyValue == "NUMPADMINUS" or keyValue == "C" then 
			if down then
				Musician.Keyboard.ShiftKeys(layer, -14)
			end
			return true
		else
			return false
		end
	end

	layer = note[1]
	noteKey = note[2]

	if key == nil then
		return false
	end

	local note = Musician.Keyboard.mapping[key]

	if note == nil then
		return true
	end

	local layer = note[1]
	local noteKey = note[2]
	local instrument = Musician.Keyboard.config.instrument[layer]
	local powerChords = Musician.Keyboard.config.powerChords[layer] and instrument < 128 -- No power chords for percussions

	if noteKey < Musician.MIN_KEY or noteKey > Musician.MAX_KEY or instrument < 0 then
		return true
	end

	if powerChords then
		Musician.Live.NoteOff(noteKey - 12, layer, instrument, true)
		Musician.Live.NoteOff(noteKey - 5, layer, instrument, true)
	end
	Musician.Live.NoteOff(noteKey, layer, instrument)

	if down then
		if powerChords then
			Musician.Live.NoteOn(noteKey - 12, layer, instrument, true, MusicianKeyboard)
			Musician.Live.NoteOn(noteKey - 5, layer, instrument, true, MusicianKeyboard)
		end
		Musician.Live.NoteOn(noteKey, layer, instrument, false, MusicianKeyboard)
	end

	return true
end

function MusicianOptions.SetSustain(enable, layer)
	if enable and sustainedNotes[layer] == nil then
		-- Create sustained notes container for layer
		sustainedNotes[layer] = {}
	elseif not(enable) and sustainedNotes[layer] ~= nil then
		-- Turn off all sustained notes
		local noteOn
		for _, noteOn in pairs(sustainedNotes[layer]) do
			local key = noteOn[NOTEON.KEY]
			local instrument = noteOn[NOTEON.INSTRUMENT]
			local isChordNote = noteOn[NOTEON.IS_CHORD_NOTE]
			-- Don't ignore sustain
			Musician.Live.NoteOff(key, layer, instrument, isChordNote, false)
		end

		-- Don't remove sustain container.
		-- This lets notes be sustained for their entire duration.
		-- New notes automatically cancel the oldest notes if we're at the limit.

		-- Remove sustained notes container
		-- sustainedNotes[layer] = nil
	end
end

-- Copy, because otherwise it doesn't work
function MusicianOptions.NoteOn(key, layer, instrument, isChordNote, source)

	-- Key is out of range
	if key < Musician.MIN_KEY or key > Musician.MAX_KEY then return end

	local soundFile, instrumentData = Musician.Sampler.GetSoundFile(instrument, key)

	local noteOnKey = key .. '-' .. layer .. '-' .. instrument

	-- This is an auto-chord note but a higher priority note actually exists: do nothing
	if isChordNote and notesOn[noteOnKey] and not(notesOn[noteOnKey][NOTEON.IS_CHORD_NOTE]) then
		return
	end

	-- Layer is sustained and note is already sustained: turn it off then start over
	if sustainedNotes[layer] ~= nil and sustainedNotes[layer][noteOnKey] ~= nil then
		local sustainedNote = sustainedNotes[layer][noteOnKey]
		Musician.Live.NoteOff(key, layer, instrument, isChordNote, true)
	end

	-- Mute game music
	muteGameMusic()

	-- Play note
	local handle = 0
	if not(Musician.Live.IsBandSyncMode() and Musician.Live.IsLiveEnabled()) then
		handle = Musician.Sampler.PlayNote(instrumentData, key)
	end

	-- Insert note on and trigger event
	Musician.Live.InsertNote(true, key, layer, instrument)
	notesOn[noteOnKey] = {
		[NOTEON.HANDLE] = handle,
		[NOTEON.LAYER] = layer,
		[NOTEON.KEY] = key,
		[NOTEON.INSTRUMENT] = instrument,
		[NOTEON.IS_CHORD_NOTE] = isChordNote,
		[NOTEON.SOURCE] = source
	}
	Musician.Live:SendMessage(Musician.Events.LiveNoteOn, key, layer, instrumentData, isChordNote, source)

	-- Send band note message if synchronization is enabled
	Musician.Live.BandNote(true, key, layer, instrument)
end
-- Copy, because otherwise it doesn't work
function MusicianOptions.NoteOff(key, layer, instrument, isChordNote, ignoreSustain)

	-- Key is out of range
	if key < Musician.MIN_KEY or key > Musician.MAX_KEY then return end

	local noteOnKey = key .. '-' .. layer .. '-' .. instrument
	if not(notesOn[noteOnKey]) then return end

	local handle = notesOn[noteOnKey][NOTEON.HANDLE]
	local noteOnIsChordNote = notesOn[noteOnKey][NOTEON.IS_CHORD_NOTE]

	-- If current note off is from an auto-chord but actual note is not: do nothing
	if isChordNote and not(noteOnIsChordNote) then
		return
	end

	-- Layer is sustained
	if not(ignoreSustain) and sustainedNotes[layer] ~= nil then
		sustainedNotes[layer][noteOnKey] = notesOn[noteOnKey]
		return
	end

	unmuteGameMusic()

	-- Stop playing note
	if handle then
		Musician.Sampler.StopNote(handle)
	end

	-- Insert note off and trigger event
	local source = notesOn[noteOnKey][NOTEON.SOURCE]
	Musician.Live.InsertNote(false, key, layer, instrument)
	notesOn[noteOnKey] = nil
	Musician.Live:SendMessage(Musician.Events.LiveNoteOff, key, layer, isChordNote, source)

	-- Send band note message if synchronization is enabled
	Musician.Live.BandNote(false, key, layer, instrument)
end
-- Copy, because otherwise it doesn't work
function MusicianOptions.AllNotesOff(onlyForLayer)
	-- Set sustain off for layers
	Musician.Utils.ForEach(sustainedNotes, function(notes, layer)
		if onlyForLayer == nil or onlyForLayer == layer then
			Musician.Live.SetSustain(false, layer)
		end
	end)

	-- Turn off notes
	local noteOnKey, note
	for noteOnKey, note in pairs(notesOn) do
		local layer = note[NOTEON.LAYER]
		local key = note[NOTEON.KEY]
		local instrument = note[NOTEON.INSTRUMENT]

		if onlyForLayer == nil or onlyForLayer == layer then
			Musician.Live.NoteOff(key, layer, instrument)
		end
	end
end

--- Return main menu elements
-- @return menu (table)
function MusicianOptions.GetMenu()
	local menu = MusicianButtonGetMenu()

	table.insert(menu, {
		notCheckable = true,
		text = MusicianOptions.Msg.MENU_TITLE,
		isTitle = true
	})

	if useGW2 then
		table.insert(menu, {
			notCheckable = true,
			text = MusicianOptions.Msg.MENU_WOW,
			func = function()
				useGW2 = false
				MusicianKeyboard.NoteKey = MusicianKeyboardNotekey
			end
		})
	else
		table.insert(menu, {
			notCheckable = true,
			text = MusicianOptions.Msg.MENU_GW2,
			func = function()
				useGW2 = true
				MusicianKeyboard.NoteKey = MusicianOptions.NoteKey
			end
		})
	end
	
	if isSustained then
		table.insert(menu, {
			notCheckable = true,
			text = MusicianOptions.Msg.MENU_SUSTAIN,
			func = function()
				isSustained = false
				Musician.Live.SetSustain = MusicianLiveSetSustain
				Musician.Live.NoteOn = MusicianLiveNoteOn 
				Musician.Live.NoteOff = MusicianLiveNoteOff
				Musician.Live.AllNotesOff = MusicianLiveAllNotesOff
			end
		})
	else
		table.insert(menu, {
			notCheckable = true,
			text = MusicianOptions.Msg.MENU_NO_SUSTAIN,
			func = function()
				isSustained = true
				Musician.Live.SetSustain = MusicianOptions.SetSustain
				Musician.Live.NoteOn = MusicianOptions.NoteOn
				Musician.Live.NoteOff = MusicianOptions.NoteOff
				Musician.Live.AllNotesOff = MusicianOptions.AllNotesOff
			end
		})
	end

	return menu
end